@@ -37,6 +37,9 @@
 
 #include "htmlcursor.h"
 
+static gboolean move_right (HTMLCursor *cursor, HTMLEngine *e);
+static gboolean move_left (HTMLCursor *cursor, HTMLEngine *e);
+
 /* #define _HTML_CURSOR_DEBUG */
 
 #ifdef _HTML_CURSOR_DEBUG
@@ -279,6 +282,7 @@ html_cursor_up (HTMLCursor *cursor,
 {
 	HTMLCursor orig_cursor;
 	HTMLCursor prev_cursor;
+	HTMLDirection dir;
 	gint prev_x, prev_y;
 	gint x, y;
 	gint target_x;
@@ -296,6 +300,11 @@ html_cursor_up (HTMLCursor *cursor,
 	if (engine->need_spell_check)
 		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
 
+	if (cursor->object->parent)
+		dir = html_object_get_direction (cursor->object->parent);
+	else
+		dir = HTML_DIRECTION_LTR;
+
 	html_cursor_copy (&orig_cursor, cursor);
 
 	html_object_get_cursor_base (cursor->object,
@@ -334,28 +343,52 @@ html_cursor_up (HTMLCursor *cursor,
 		if (y + cursor->object->descent - 1 < prev_y - prev_cursor.object->ascent) {
 			if (new_line) {
 				html_cursor_copy (cursor, &prev_cursor);
-				return FALSE;
+				return TRUE;
 			}
 
 			new_line = TRUE;
+			if (cursor->object->parent)
+				dir = html_object_get_direction (cursor->object->parent);
+			else
+				dir = HTML_DIRECTION_LTR;
 		}
 
-		if (new_line && x <= target_x) {
-			if (! cursor->have_target_x) {
-				cursor->have_target_x = TRUE;
-				cursor->target_x = target_x;
-			}
+		if (dir == HTML_DIRECTION_RTL) {
+			if (new_line && x >= target_x) {
+				if (! cursor->have_target_x) {
+					cursor->have_target_x = TRUE;
+					cursor->target_x = target_x;
+				}
 
-			/* Choose the character which is the nearest to the
-                           target X.  */
-			if (prev_y == y && target_x - x >= prev_x - target_x) {
-				cursor->object = prev_cursor.object;
-				cursor->offset = prev_cursor.offset;
-				cursor->position = prev_cursor.position;
-			}
+				/* Choose the character which is the nearest to the
+				   target X.  */
+				if (prev_y == y && x - target_x >= target_x - prev_x) {
+					cursor->object = prev_cursor.object;
+					cursor->offset = prev_cursor.offset;
+					cursor->position = prev_cursor.position;
+				}
 
-			debug_location (cursor);
-			return TRUE;
+				debug_location (cursor);
+				return TRUE;
+			}
+		} else {
+			if (new_line && x <= target_x) {
+				if (! cursor->have_target_x) {
+					cursor->have_target_x = TRUE;
+					cursor->target_x = target_x;
+				}
+
+				/* Choose the character which is the nearest to the
+				   target X.  */
+				if (prev_y == y && target_x - x >= prev_x - target_x) {
+					cursor->object = prev_cursor.object;
+					cursor->offset = prev_cursor.offset;
+					cursor->position = prev_cursor.position;
+				}
+
+				debug_location (cursor);
+				return TRUE;
+			}
 		}
 	}
 }
@@ -367,6 +400,7 @@ html_cursor_down (HTMLCursor *cursor,
 {
 	HTMLCursor orig_cursor;
 	HTMLCursor prev_cursor;
+	HTMLDirection dir;
 	gint prev_x, prev_y;
 	gint x, y;
 	gint target_x;
@@ -384,6 +418,13 @@ html_cursor_down (HTMLCursor *cursor,
 	if (engine->need_spell_check)
 		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
 
+	if (cursor->object->parent)
+		dir = html_object_get_direction (cursor->object->parent);
+	else
+		dir = HTML_DIRECTION_LTR;
+
+	html_cursor_copy (&orig_cursor, cursor);
+
 	html_object_get_cursor_base (cursor->object,
 				     engine->painter, cursor->offset,
 				     &x, &y);
@@ -404,8 +445,13 @@ html_cursor_down (HTMLCursor *cursor,
 		prev_x = x;
 		prev_y = y;
 
-		if (! forward (cursor))
-			return FALSE;
+		if (dir == HTML_DIRECTION_RTL) {
+			if (! move_left (cursor, engine))
+				return FALSE;
+		} else {
+			if (! move_right (cursor, engine))
+				return FALSE;
+		}
 
 		html_object_get_cursor_base (cursor->object,
 					     engine->painter, cursor->offset,
@@ -419,28 +465,52 @@ html_cursor_down (HTMLCursor *cursor,
 		if (y - cursor->object->ascent > prev_y + prev_cursor.object->descent - 1) {
 			if (new_line) {
 				html_cursor_copy (cursor, &prev_cursor);
-				return FALSE;
+				return TRUE;
 			}
 
 			new_line = TRUE;
+			if (cursor->object->parent)
+				dir = html_object_get_direction (cursor->object->parent);
+			else
+				dir = HTML_DIRECTION_LTR;
 		}
 
-		if (new_line && x >= target_x) {
-			if (! cursor->have_target_x) {
-				cursor->have_target_x = TRUE;
-				cursor->target_x = target_x;
-			}
+		if (dir == HTML_DIRECTION_RTL) {
+			if (new_line && x <= target_x) {
+				if (! cursor->have_target_x) {
+					cursor->have_target_x = TRUE;
+					cursor->target_x = target_x;
+				}
 
-			/* Choose the character which is the nearest to the
-                           target X.  */
-			if (prev_y == y && x - target_x >= target_x - prev_x) {
-				cursor->object = prev_cursor.object;
-				cursor->offset = prev_cursor.offset;
-				cursor->position = prev_cursor.position;
-			}
+				/* Choose the character which is the nearest to the
+				   target X.  */
+				if (prev_y == y && target_x - x >= prev_x - target_x) {
+					cursor->object = prev_cursor.object;
+					cursor->offset = prev_cursor.offset;
+					cursor->position = prev_cursor.position;
+				}
 
-			debug_location (cursor);
-			return TRUE;
+				debug_location (cursor);
+				return TRUE;
+			}
+		} else {
+			if (new_line && x >= target_x) {
+				if (! cursor->have_target_x) {
+					cursor->have_target_x = TRUE;
+					cursor->target_x = target_x;
+				}
+
+				/* Choose the character which is the nearest to the
+				   target X.  */
+				if (prev_y == y && x - target_x >= target_x - prev_x) {
+					cursor->object = prev_cursor.object;
+					cursor->offset = prev_cursor.offset;
+					cursor->position = prev_cursor.position;
+				}
+
+				debug_location (cursor);
+				return TRUE;
+			}
 		}
 	}
 }
@@ -773,7 +843,7 @@ html_cursor_child_of (HTMLCursor *cursor, HTMLObject *parent)
 }
 
 static gboolean
-move_to_next_object (HTMLCursor *cursor)
+move_to_next_object (HTMLCursor *cursor, HTMLEngine *e)
 {
 	HTMLObject *next;
 
@@ -782,12 +852,14 @@ move_to_next_object (HTMLCursor *cursor)
 		cursor->object = next;
 		cursor->position ++;
 
-		if (html_object_get_direction (next->parent) == HTML_DIRECTION_RTL) {
-			cursor->offset = html_object_get_right_edge_offset (next, 0);
-		} else {
-			cursor->offset = html_object_get_left_edge_offset (next, 0);
+		if (!html_object_is_container (next)) {
+			if (html_object_get_direction (next->parent) == HTML_DIRECTION_RTL) {
+				cursor->offset = html_object_get_right_edge_offset (next, e->painter, 0);
+			} else {
+				cursor->offset = html_object_get_left_edge_offset (next, e->painter, 0);
+			}
+			cursor->position += cursor->offset;
 		}
-		cursor->position += cursor->offset;
 
 		return TRUE;
 	} else
@@ -795,7 +867,7 @@ move_to_next_object (HTMLCursor *cursor)
 }
 
 static gboolean
-move_to_prev_object (HTMLCursor *cursor)
+move_to_prev_object (HTMLCursor *cursor, HTMLEngine *e)
 {
 	HTMLObject *prev;
 
@@ -804,12 +876,14 @@ move_to_prev_object (HTMLCursor *cursor)
 		cursor->object = prev;
 		cursor->position --;
 
-		if (html_object_get_direction (prev->parent) == HTML_DIRECTION_RTL) {
-			cursor->offset = html_object_get_left_edge_offset (prev, html_object_get_length (prev));
-		} else {
-			cursor->offset = html_object_get_right_edge_offset (prev, html_object_get_length (prev));
+		if (!html_object_is_container (prev)) {
+			if (html_object_get_direction (prev->parent) == HTML_DIRECTION_RTL) {
+				cursor->offset = html_object_get_left_edge_offset (prev, e->painter, html_object_get_length (prev));
+			} else {
+				cursor->offset = html_object_get_right_edge_offset (prev, e->painter, html_object_get_length (prev));
+			}
+			cursor->position -= cursor->offset - html_object_get_length (prev);
 		}
-		cursor->position -= cursor->offset - html_object_get_length (prev);
 
 		return TRUE;
 	} else
@@ -817,14 +891,14 @@ move_to_prev_object (HTMLCursor *cursor)
 }
 
 static gboolean
-move_left (HTMLCursor *cursor)
+move_left (HTMLCursor *cursor, HTMLEngine *e)
 {
-	if (!html_object_cursor_left (cursor->object, cursor)) {
+	if (!html_object_cursor_left (cursor->object, e->painter, cursor)) {
 		if (cursor->object->parent) {
 			if (html_object_get_direction (cursor->object->parent) == HTML_DIRECTION_RTL)
-				return move_to_next_object (cursor);
+				return move_to_next_object (cursor, e);
 			else
-				return move_to_prev_object (cursor);
+				return move_to_prev_object (cursor, e);
 		}
 	}
 
@@ -845,7 +919,7 @@ html_cursor_left (HTMLCursor *cursor, HTMLEngine *engine)
 		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
 
 	cursor->have_target_x = FALSE;
-	retval = move_left (cursor);
+	retval = move_left (cursor, engine);
 
 	debug_location (cursor);
 
@@ -853,22 +927,22 @@ html_cursor_left (HTMLCursor *cursor, HTMLEngine *engine)
 }
 
 static gboolean
-left_in_flow (HTMLCursor *cursor)
+left_in_flow (HTMLCursor *cursor, HTMLEngine *e)
 {
 	gboolean retval;
 
 	retval = TRUE;
-	if (cursor->offset != html_object_get_left_edge_offset (cursor->object, cursor->offset) && html_object_is_container (cursor->object)) {
+	if (cursor->offset != html_object_get_left_edge_offset (cursor->object, e->painter, cursor->offset) && html_object_is_container (cursor->object)) {
 		HTMLObject *obj;
 
 		obj = cursor->object;
-		while ((retval = move_left (cursor)) && cursor->object != obj)
+		while ((retval = move_left (cursor, e)) && cursor->object != obj)
 			;
 	} else {
 		if (cursor->offset > 1 || !cursor->object->prev)
-			retval = html_object_cursor_left (cursor->object, cursor);
+			retval = html_object_cursor_left (cursor->object, e->painter, cursor);
 		else if (cursor->object->prev)
-			retval = move_left (cursor);
+			retval = move_left (cursor, e);
 		else
 			retval = FALSE;
 	}
@@ -898,7 +972,7 @@ html_cursor_left_edge_of_line (HTMLCursor *cursor, HTMLEngine *engine)
 				     &x, &prev_y);
 
 	while (1) {
-		if (! left_in_flow (cursor))
+		if (! left_in_flow (cursor, engine))
 			return TRUE;
 
 		html_object_get_cursor_base (cursor->object, engine->painter, cursor->offset,
@@ -915,17 +989,17 @@ html_cursor_left_edge_of_line (HTMLCursor *cursor, HTMLEngine *engine)
 }
 
 static gboolean
-move_right (HTMLCursor *cursor)
+move_right (HTMLCursor *cursor, HTMLEngine *e)
 {
 	gboolean retval;
 
 	retval = TRUE;
-	if (!html_object_cursor_right (cursor->object, cursor)) {
+	if (!html_object_cursor_right (cursor->object, e->painter, cursor)) {
 		if (cursor->object->parent) {
 			if (html_object_get_direction (cursor->object->parent) == HTML_DIRECTION_RTL)
-				return move_to_prev_object (cursor);
+				return move_to_prev_object (cursor, e);
 			else
-				return move_to_next_object (cursor);
+				return move_to_next_object (cursor, e);
 		}
 	}
 	return retval;
@@ -945,7 +1019,7 @@ html_cursor_right (HTMLCursor *cursor, HTMLEngine *engine)
 		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
 
 	cursor->have_target_x = FALSE;
-	retval = move_right (cursor);
+	retval = move_right (cursor, engine);
 
 	debug_location (cursor);
 
@@ -953,23 +1027,23 @@ html_cursor_right (HTMLCursor *cursor, HTMLEngine *engine)
 }
 
 static gboolean
-right_in_flow (HTMLCursor *cursor)
+right_in_flow (HTMLCursor *cursor, HTMLEngine *e)
 {
 	gboolean retval;
 
 	retval = TRUE;
-	if (cursor->offset != html_object_get_right_edge_offset (cursor->object, cursor->offset)) {
+	if (cursor->offset != html_object_get_right_edge_offset (cursor->object, e->painter, cursor->offset)) {
 		if (html_object_is_container (cursor->object)) {
 			HTMLObject *obj;
 
 			obj = cursor->object;
-			while ((retval = move_right (cursor)) && cursor->object != obj)
+			while ((retval = move_right (cursor, e)) && cursor->object != obj)
 				;
 		} else
-			retval = html_object_cursor_right (cursor->object, cursor);
+			retval = html_object_cursor_right (cursor->object, e->painter, cursor);
 	} else {
 		if (html_object_next_not_slave (cursor->object))
-			retval = move_right (cursor);
+			retval = move_right (cursor, e);
 		else
 			retval = FALSE;
 	}
@@ -999,7 +1073,7 @@ html_cursor_right_edge_of_line (HTMLCursor *cursor, HTMLEngine *engine)
 				     &x, &prev_y);
 
 	while (1) {
-		if (! right_in_flow (cursor))
+		if (! right_in_flow (cursor, engine))
 			return TRUE;
 
 		html_object_get_cursor_base (cursor->object, engine->painter, cursor->offset,
